#pragma once

/*
 * File:   thread_pool.hpp
 * Author: regan
 *
 */

#include <functional>
#include <memory>
#include <thread>

#include <upcxx/upcxx.hpp>

#include "upcxx_utils/log.hpp"
#include "upcxx_utils/timers.hpp"

// define this to disable all extra thread workers and any calls to enqueue_* will execute immediately
// #define UPCXX_UTILS_NO_THREAD_POOL

namespace upcxx_utils {

class ThreadPool_detail;
class ThreadPool {
  //
  // ThreadPool is based mostly on the ThreadPool.h from https://github.com/progschj/ThreadPool
  // by Jakob Progsch, Václav Zeman.
  //
  // This class was rewritten in 2020 by Rob Egan for use within upcxx and upcxx_utils
  //

  //
  //  Copyright (c) 2012 Jakob Progsch, Václav Zeman
  //
  // This software is provided 'as-is', without any express or implied
  // warranty. In no event will the authors be held liable for any damages
  // arising from the use of this software.
  //
  // Permission is granted to anyone to use this software for any purpose,
  // including commercial applications, and to alter it and redistribute it
  // freely, subject to the following restrictions:
  //
  // 1. The origin of this software must not be misrepresented; you must not
  //    claim that you wrote the original software. If you use this software
  //    in a product, an acknowledgment in the product documentation would be
  //    appreciated but is not required.
  //
  // 2. Altered source versions must be plainly marked as such, and must not be
  //    misrepresented as being the original software.
  //
  // 3. This notice may not be removed or altered from any source
  //    distribution.

  using Task = std::function<void()>;  // void(void) function/lambda wrappers
  std::unique_ptr<ThreadPool_detail> tp_detail;

  bool is_ready() const;
  bool is_terminal() const;

  void enqueue_task(std::shared_ptr<Task> sh_task);

 public:
  static size_t &global_task_id() {
    static size_t task_id = 0;
    return task_id;
  }
  static ThreadPool &get_single_pool(int num_threads);
  static void join_single_pool();

  template <typename Func, class... Args>
  static auto enqueue_in_single_pool(int num_workers, Func &&func, Args &&... args) {
    auto &tp = ThreadPool::get_single_pool(num_workers);
    return tp.enqueue(std::forward<Func>(func), std::forward<Args>(args)...);
  }

  ThreadPool(int num_threads);
  ThreadPool(const ThreadPool &copy) = delete;
  ThreadPool(ThreadPool &&move) = delete;
  ThreadPool &operator=(const ThreadPool &copy) = delete;
  ThreadPool &&operator=(ThreadPool &move) = delete;
  ~ThreadPool();

  template <class Func, class... Args>
  auto enqueue_with_return(Func &&func, Args &&... args) -> upcxx::future<typename std::invoke_result<Func, Args...>::type> {
#ifdef UPCXX_UTILS_NO_THREAD_POOL
    // execute immediately
    return make_future(func(args...));
#endif
    using return_t = typename std::invoke_result<Func, Args...>::type;
    static_assert(!std::is_void<return_t>::value,
                  "void is not a valid return type for enqueue_with_return... did you mean enqueue()?");
    upcxx::persona &persona = upcxx::current_persona();
    std::shared_ptr<upcxx::promise<return_t> > sh_prom = std::make_shared<upcxx::promise<return_t> >();
    sh_prom->require_anonymous(1);  // additional requirement to complete

    auto task_id = global_task_id()++;
    auto start_t = Timer::now();
    DBG("sh_prom=", sh_prom.get(), " task_id=", task_id, "\n");

    auto args_tuple = std::forward_as_tuple(args...);
    auto task =
        std::make_shared<Task>([task_id, start_t, sh_prom, &persona, func{std::move(func)}, args_tuple{std::move(args_tuple)}]() {
          DBG_VERBOSE("Executing sh_prom=", sh_prom.get(), "\n");
          sh_prom->fulfill_result(std::apply(func, args_tuple));
          DBG_VERBOSE("Finished sh_prom=", sh_prom.get(), "\n");
          // fulfill only in calling persona
          persona.lpc_ff([task_id, start_t, sh_prom]() {
            duration_seconds s = Timer::now() - start_t;
            DBG("Fulfilled sh_prom=", sh_prom.get(), " task_id=", task_id, " in ", s.count(), " s\n");
            sh_prom->fulfill_anonymous(1);
          });
        });
    enqueue_task(task);
    return sh_prom->get_future();
  }

  template <class Func, class... Args>
  upcxx::future<> enqueue_no_return(Func &&func, Args &&... args) {
#ifdef UPCXX_UTILS_NO_THREAD_POOL
    // execute immediately
    func(args...);
    return make_future();
#endif
    using return_t = typename std::invoke_result<Func, Args...>::type;
    static_assert(std::is_void<return_t>::value,
                  "void is the required return type for enqueue... did you mean enqueue_with_return()?");
    upcxx::persona &persona = upcxx::current_persona();
    std::shared_ptr<upcxx::promise<> > sh_prom = std::make_shared<upcxx::promise<> >();

    auto task_id = global_task_id()++;
    auto start_t = Timer::now();
    DBG("sh_prom=", sh_prom.get(), " task_id=", task_id, "\n");

    auto args_tuple = std::forward_as_tuple(args...);
    auto sh_task =
        std::make_shared<Task>([sh_prom, task_id, start_t, &persona, func{std::move(func)}, args_tuple{std::move(args_tuple)}]() {
          DBG_VERBOSE("Executing sh_prom=", sh_prom.get(), "\n");
          std::apply(func, args_tuple);
          DBG_VERBOSE("Finished sh_prom=", sh_prom.get(), "\n");
          // fulfill only in calling persona
          persona.lpc_ff([task_id, start_t, sh_prom]() {
            duration_seconds s = Timer::now() - start_t;
            DBG("Fulfilled sh_prom=", sh_prom.get(), " task_id=", task_id, " in ", s.count(), " s\n");
            sh_prom->fulfill_anonymous(1);
          });
        });
    enqueue_task(sh_task);
    return sh_prom->get_future();
  }

  template <class Func, class... Args>
  auto enqueue(Func &&func, Args &&... args) {
    using result_t = typename std::invoke_result<Func, Args...>::type;
    if constexpr (std::is_void<result_t>::value)
      return enqueue_no_return(std::forward<Func>(func), std::forward<Args>(args)...);
    else
      return enqueue_with_return(std::forward<Func>(func), std::forward<Args>(args)...);
  }

  void join_workers();

  void reset(int num_workers);
};

template <typename Func, class... Args>
auto execute_in_thread_pool(int num_workers, Func &&func, Args &&... args) {
  return ThreadPool::enqueue_in_single_pool(num_workers, std::forward<Func>(func), std::forward<Args>(args)...);
};

template <typename Func, class... Args>
auto execute_in_thread_pool(Func &&func, Args &&... args) {
  return execute_in_thread_pool(1, std::forward<Func>(func), std::forward<Args>(args)...);
};

template <typename Func, class... Args>
auto execute_in_new_thread(Func &&func, Args &&... args) {
  using result_t = typename std::invoke_result<Func, Args...>::type;
#ifdef UPCXX_UTILS_NO_THREAD_POOL
  // execute immediately
  if constexpr (std::is_void<result_t>::value) {
    return func(args...);
  } else {
    fund(args...);
    return;
  }
#endif
  // Create a new temporary single threaded thread pool with lifetime of the task
  auto sh_tp = make_shared<ThreadPool>(1);
  if constexpr (std::is_void<result_t>::value) {
    upcxx::future<> fut = sh_tp->enqueue_no_return(std::forward<Func>(func), std::forward<Args>(args)...);
    return fut.then([sh_tp]() {});
  } else {
    upcxx::future<result_t> fut = sh_tp->enqueue_with_return(std::forward<Func>(func), std::forward<Args>(args)...);
    return fut.then([sh_tp](result_t &&res) { return res; });
  }
}

/*  Old code

// Func no argument returned or given lambda - void()

template <typename Func>
upcxx::future<> execute_in_new_thread(upcxx::persona &persona, Func func) {
  assert(persona.active_with_caller());
  std::shared_ptr<upcxx::promise<> > sh_prom = std::make_shared<upcxx::promise<> >();

  std::shared_ptr<std::thread> sh_run = std::make_shared<std::thread>([&persona, func, sh_prom] {
    func();

    // fulfill only in calling persona
    persona.lpc_ff([&persona, sh_prom]() {
      assert(persona.active_with_caller());
      sh_prom->fulfill_anonymous(1);
    });
  });
  auto fut_finished = sh_prom->get_future().then(  // keep sh_prom and sh_run alive until complete
      [&persona, sh_prom, sh_run]() {
        assert(persona.active_with_caller());
        sh_run->join();
      });
  return fut_finished;
}

template <typename Func>
upcxx::future<> execute_in_new_thread(Func func) {
  return execute_in_new_thread(upcxx::current_persona(), func);
}
*/

};  // namespace upcxx_utils
