#include "upcxx_utils/thread_pool.hpp"

//
// ThreadPool is based mostly on the ThreadPool.h from https://github.com/progschj/ThreadPool
// by Jakob Progsch, Václav Zeman.
//
// This class was rewritten in 2020 by Rob Egan for use within upcxx and upcxx_utils
//

//
//  Copyright (c) 2012 Jakob Progsch, Václav Zeman
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
//
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
//
// 3. This notice may not be removed or altered from any source
//    distribution.

#include <atomic>
#include <condition_variable>
#include <functional>
#include <memory>
#include <mutex>
#include <queue>
#include <thread>
#include <type_traits>
#include <upcxx/upcxx.hpp>
#include <vector>

class upcxx_utils::ThreadPool_detail {
 public:
  // the threads
  std::vector<std::thread> workers;
  std::vector<IntermittentTimer> worker_timers;

  // the task queue
  using Task = std::function<void()>;  // void(void) function/lambda wrappers
  using Mutex = std::mutex;
  using Lock = std::unique_lock<Mutex>;
  std::queue<Task> tasks;
  Mutex task_mutex, workers_mutex;
  std::condition_variable task_ready;
  int max_workers;
  std::atomic<int> ready_workers;
  bool stop;

  ThreadPool_detail()
      : workers()
      , tasks()
      , task_mutex()
      , workers_mutex()
      , task_ready()
      , max_workers(0)
      , ready_workers(0)
      , stop(false) {}

  ~ThreadPool_detail() { join_workers(); }

  bool is_ready() const { return stop | !tasks.empty(); }
  bool is_terminal() const { return stop & tasks.empty(); }

  void set_max_workers(int new_max_workers) {
#ifdef UPCXX_UTILS_NO_THREAD_POOL
    max_workers = 0;
    return;  // spawn no threads
#endif
    if (max_workers > new_max_workers) return;  // Cannot shrink the pool presently
    if (new_max_workers != max_workers) {
      Lock lock(workers_mutex);
      max_workers = new_max_workers;
      workers.reserve(max_workers);
      worker_timers.reserve(max_workers);
    }
  }

  void enqueue_task(std::shared_ptr<Task> sh_task) {
    if (stop) {
      DIE("Cannot enqueue a task to a stopped ThreadPool: stop=", stop, " max_workers=", max_workers);
    }
    if (max_workers == 0 && workers.empty()) {
      // execute immediately
      (*sh_task)();
      return;
    }
    {
      Lock lock(task_mutex);
      if (stop) DIE("Can not enqueue to a stopped ThreadPool");
      tasks.emplace([sh_task]() { (*sh_task)(); });
    }
    // lazy construct new threads for the pool
    if (workers.size() < max_workers && ready_workers.load() == 0) {
      add_one_worker();
    }
    task_ready.notify_one();
  }

  void join_workers() {
    DBG("Joining ", workers.size(), " workers\n");
    if (workers.empty()) return;
    // set stop
    {
      Lock lock(this->task_mutex);
      stop = true;
    }
    // notify all
    task_ready.notify_all();
    // join all
    for (auto &worker : workers) worker.join();
    assert(ready_workers.load() == 0 && "All workers completed");
    Lock lock(workers_mutex);
    workers.clear();
    for (auto &timer : worker_timers) {
      LOG("Worker joined after executing ", timer.get_count(), " tasks over ", timer.get_elapsed(), " s.\n");
    }
    worker_timers.clear();
  }

  void reset(int num_workers) {
    join_workers();
    assert(stop);
    stop = false;
    set_max_workers(num_workers);
  }

 protected:
  void add_one_worker() {
    if (stop) DIE("Cannot add workers to a stopped ThreadPool");
    assert(workers.size() < max_workers);
    Lock lock(workers_mutex);  // modifying workers, so lock
    worker_timers.emplace_back("ThreadWorkerTimer");
    workers.emplace_back([this, &timer = worker_timers.back()] {
      DBG_VERBOSE(std::this_thread::get_id(), " just started\n");
      while (true) {
        Task task;
        {
          Lock lock(this->task_mutex);
          this->ready_workers++;
          this->task_ready.wait(lock, [this] { return this->is_ready(); });
          this->ready_workers--;
          if (this->is_terminal()) {
            DBG_VERBOSE(std::this_thread::get_id(), " terminated\n");
            return;
          }

          // pop the next task
          assert(!this->tasks.empty());
          task = std::move(this->tasks.front());
          this->tasks.pop();
        }
        DBG_VERBOSE(std::this_thread::get_id(), " popped task# ", timer.get_count(), "\n");
        timer.start();
        task();  // execute
        DBG_VERBOSE(std::this_thread::get_id(), " done with task# ", timer.get_count(), " in ", timer.get_elapsed_since_start(),
                    " s\n");
        timer.stop();
      }
    });
    DBG_VERBOSE("Added a new worker: ", workers.back().get_id(), "\n");
  }
};

bool upcxx_utils::ThreadPool::is_ready() const { return tp_detail->is_ready(); }
bool upcxx_utils::ThreadPool::is_terminal() const { return tp_detail->is_terminal(); }
void upcxx_utils::ThreadPool::enqueue_task(std::shared_ptr<Task> sh_task) { tp_detail->enqueue_task(sh_task); }

upcxx_utils::ThreadPool &upcxx_utils::ThreadPool::get_single_pool(int num_threads) {
  static ThreadPool _the_singleton_pool_(num_threads);
  _the_singleton_pool_.tp_detail->set_max_workers(num_threads);
  return _the_singleton_pool_;
}
void upcxx_utils::ThreadPool::join_single_pool() { get_single_pool(0).join_workers(); }

upcxx_utils::ThreadPool::ThreadPool(int num_threads)
    : tp_detail(std::make_unique<ThreadPool_detail>()) {
  // reserve and start the threads in the pool
  tp_detail->set_max_workers(num_threads);
  if (num_threads == 0) LOG("NOTICE: A ThreadPool with 0 threads just runs tasks in the calling thread immediately\n");
  DBG("Constructed with ", num_threads, "\n");
}

upcxx_utils::ThreadPool::~ThreadPool() { join_workers(); }

void upcxx_utils::ThreadPool::join_workers() { tp_detail->join_workers(); }

void upcxx_utils::ThreadPool::reset(int num_workers) { tp_detail->reset(num_workers); }
