#include <unistd.h>

#include <atomic>
#include <iostream>
#include <string>
#include <upcxx/upcxx.hpp>

#include "upcxx_utils/thread_pool.hpp"
#include "upcxx_utils/timers.hpp"
#include "upcxx_utils/version.h"

using upcxx_utils::ThreadPool;
using upcxx::future;
using upcxx::promise;

void test_threads(int num_threads) {
  future<> all_done = make_future();
  {
    // construct destruct
    ThreadPool tp(num_threads);
  }

  int iterations = 250;
  std::atomic<size_t> sum(0);
  {
    // construct send jobs and join
    ThreadPool tp(num_threads);
    for (int i = 0; i < iterations; i++) {
      auto fut = tp.enqueue([&sum, i, num_threads]() {
        if (num_threads == 1) assert(i == sum.load() && "one thread is serial");
        sum++;
      });
      all_done = when_all(all_done, fut);
    }
    all_done.wait();
    assert(sum.load() == iterations && "all tasks completed after wait");
  }
  assert(all_done.ready());

  sum.store(0);
  {
    // construct send jobs and join
    ThreadPool tp(num_threads);
    for (int i = 0; i < iterations; i++) {
      auto fut = tp.enqueue([&sum, i, num_threads]() {
        if (num_threads == 1) assert(i == sum.load() && "one thread is serial");
        sum++;
      });
      all_done = when_all(all_done, fut);
    }
    // no wait necessary here
  }
  // all tasks are complete.
  assert(sum.load() == iterations && "all tasks completed after deconstruct");
  all_done.wait();  // futures may not be ready, however

  // enqueue with return
  sum.store(0);
  {
    // construct send jobs and join
    ThreadPool tp(num_threads);
    for (int i = 0; i < iterations; i++) {
      auto fut = tp.enqueue([&sum, i, num_threads]() {
                     if (num_threads == 1) assert(i == sum.load() && "one thread is serial");
                     sum++;
                     return i;
                   }).then([i](int v) { assert(i == v && "return v == i"); });
      all_done = when_all(all_done, fut);
    }
    all_done.wait();
    assert(sum.load() == iterations && "all tasks completed after wait");
  }
  assert(all_done.ready());

  sum.store(0);
  {
    // construct send jobs and join
    ThreadPool tp(num_threads);
    for (int i = 0; i < iterations; i++) {
      auto fut = tp.enqueue([&sum, i, num_threads]() {
                     if (num_threads == 1) assert(i == sum.load() && "one thread is serial");
                     sum++;
                     return i;
                   }).then([i](int v) { assert(i == v && "return v == i"); });
      all_done = when_all(all_done, fut);
    }
    // no wait necessary here
  }
  // all tasks are complete.
  assert(sum.load() == iterations && "all tasks completed after destruct");
  all_done.wait();  // futures may not be however

  sum.store(0);
  {
    promise prom1(1);
    // construct within progress callback serially send jobs and join
    ThreadPool tp(num_threads);
    for (int i = 0; i < iterations; i++) {
      all_done = when_all(all_done, prom1.get_future()).then([&sum, &tp, i, num_threads]() {
        return tp.enqueue([&sum, i, num_threads]() {
          if (num_threads == 1) assert(i == sum.load() && "one thread is serial");
          sum++;
        });
      });
    }
    prom1.fulfill_anonymous(1);
    all_done.wait();
    assert(sum.load() == iterations && "all tasks completed after wait");
  }

  sum.store(0);
  {
    promise prom1(1);
    // construct within progress callback
    ThreadPool tp(num_threads);
    for (int i = 0; i < iterations; i++) {
      auto fut = prom1.get_future().then([&sum, &tp, i, num_threads]() { return tp.enqueue([&sum, i, num_threads]() { sum++; }); });
      all_done = when_all(all_done, fut);
    }
    prom1.fulfill_anonymous(1);
    all_done.wait();
    assert(sum.load() == iterations && "all tasks completed after wait");
  }
}

int test_thread_pool(int argc, char **argv) {
  if (!upcxx::rank_me()) std::cout << "Found upcxx_utils version " << UPCXX_UTILS_VERSION << std::endl;
  char hnbuf[64];
  gethostname(hnbuf, sizeof(hnbuf) - 1);
  if (upcxx::local_team().rank_me() == 0) {
    std::cout << "proc " << upcxx::rank_me() << " on " << hnbuf << std::endl;
  }
  upcxx_utils::open_dbg("test_thread_pool");

  for (int threads = 0; threads <= 10; threads++) {
    barrier();
    SOUT("Running tests with ", threads, " thread(s) in the pool... ");
    upcxx_utils::BaseTimer t("");
    t.start();
    test_threads(threads);
    barrier();
    t.stop();
    SOUT(t.get_elapsed(), " s.\n");
  }
  barrier();

  upcxx_utils::close_dbg();
  return 0;
}
